// PING_PONG SCOREBOARD v2.4

// Include libraries for MAX72XX display
// Make sure these install these libraries before compiling
// > Arduino > Tools > Manage Libraries 
#include <SPI.h> // I believe this library is built-in or installed by default
#include <MD_MAX72xx.h>
#include <MD_Parola.h>

// Include custom font files
#include "fontSlim3x7.h"
#include "fontTallNumTop.h"
#include "fontTallNumBot.h"

#define DEBUG 1

// ==================================================================
// == LED DISPLAY ===================================================
// ==================================================================

// Define hardware type, size, and output pins:
#define HARDWARE_TYPE MD_MAX72XX::FC16_HW
#define MAX_ZONES 2
#define ZONE_SIZE 4
#define MAX_DEVICES (MAX_ZONES * ZONE_SIZE)

// Define which pins you have plugged your display in to
#define CLK_PIN   13
#define DATA_PIN  11
#define CS_PIN    3

#define ZONE_UPPER  1
#define ZONE_LOWER  0

// Create a new instance of the MD_Parola class with hardware SPI connection:
MD_Parola P = MD_Parola(HARDWARE_TYPE, CS_PIN, MAX_DEVICES);

// Global variables
uint8_t  curMsg;
const char *msg[] =
{
  "THE SHOP",
  "SCOREBRD",
  "PRESS",
  "TO BEGIN",
  "FIRST TO",
  "21 WINS!",
  "WINNER!"
};

// Default text display parameters
textEffect_t scrollEffectIn = PA_SCROLL_LEFT;
textEffect_t scrollEffectOut = PA_SCROLL_LEFT;
textPosition_t scrollAlign = PA_LEFT;
uint16_t scrollSpeed = 25; // in milliseconds
uint16_t scrollPause = 2000; // in milliseconds

// Input pin for sound buzzer
int buzzer = 4;

// Global message buffers shared by Scrolling functions
#define  BUF_SIZE  75
char curMessage[BUF_SIZE] = {""};
char newMessage[BUF_SIZE] = {""};
bool newMessageAvailable = true;

// ==================================================================
// == GAME CONFIG ===================================================
// ==================================================================

const int player1pin = 7;     // number of input pin for Button 1 
const int player2pin = 6;     // number of input pin for Button 2 
const int ledPin =  13;       // the number of the LED pin
const int DEBOUNCE_DELAY = 50; // 50ms between allowed button presses
const int NUMBER_OF_PLAYERS = 2;

const int TRUE = 1;
const int FALSE = 0;

const int SCOREBOARD_INIT = 0;
const int GAME_WAITING    = 1;
const int GAME_START      = 2;
const int GAME_ON         = 3;
const int GAME_OVER       = 4;

int game_state = SCOREBOARD_INIT;

// States of the button
int UNPRESSED = 1;
int PRESSED = 2;

typedef struct ScoreBoard {
  int p1score;
  int p2score;
} ScoreBoard;

ScoreBoard scoreboard;

typedef struct Player {
  int pin;
  int buttonInputState;
  int buttonState;
  int score;
  unsigned long lastDebounceTime;  // the last time the output pin was toggled
  unsigned long debounceDelay;    // the debounce time; increase if the output flickers
} Player;

Player player1;
Player player2;

// Allocate memory for two Player structs.
Player players[2];


// ===================================================================
// == SETUP ==========================================================
// ===================================================================
void setup() {

#if DEBUG
  Serial.begin(57600);
  Serial.print("[Ping Pong Scoreboard v2 init...]\n");
#endif

  // Initialize I/O Pins
  pinMode(ledPin, OUTPUT);    // initialize the LED pin as an output:
  pinMode(player1pin, INPUT); // initialize the pushbutton pin as an input:
  pinMode(player2pin, INPUT); // initialize the pushbutton pin as an input:
  pinMode(buzzer, OUTPUT);    // initialize the buzzer pin as an input:

  digitalWrite(player1pin, HIGH);
  digitalWrite(player2pin, HIGH);

  for (int i = 0; i < NUMBER_OF_PLAYERS; i++) {
    players[i].buttonInputState = LOW;
    players[i].buttonState = UNPRESSED;
    players[i].score = 0;
    players[i].lastDebounceTime = 0;  // the last time the output pin was toggled
  }

  scoreboard.p1score = -1;
  scoreboard.p2score = -1;

  players[0].pin = player1pin;
  players[1].pin = player2pin;

  // Init LED display
  P.begin(MAX_ZONES);

  // Define upper and lower displays, and reset to default font
  P.setZone(ZONE_LOWER, 0, ZONE_SIZE - 1);
  P.setFont(ZONE_LOWER, NULL);
  P.setZone(ZONE_UPPER, ZONE_SIZE, MAX_DEVICES-1);
  P.setFont(ZONE_UPPER, NULL);

  // Set display brightness to max
  P.setIntensity(15);  // Set the intensity (brightness) of the display (0-15):
  P.displayClear();   // Clear the display:
}

// ===================================================================
// == INIT ===========================================================
// ===================================================================
int scoreboard_init() {

  // Set upper and lower zones to single-high font for this message
  P.setFont(ZONE_UPPER, fontSlim);
  P.setFont(ZONE_LOWER, fontSlim);
  P.setCharSpacing(P.getCharSpacing() * 1); // single height

  // Upper zone; "THE SHOP",
  // Lower zone: "SCOREBRD",
  P.displayZoneText(ZONE_UPPER, msg[0], PA_CENTER, 0, 0, PA_PRINT, PA_NO_EFFECT);
  P.displayZoneText(ZONE_LOWER, msg[1], PA_CENTER, 0, 0, PA_PRINT, PA_NO_EFFECT);

  // Loop for 8 seconds to display message
  unsigned long tmr = millis();
  while(true){
    if (P.displayAnimate()) {
      P.displayReset();
    }
    if (millis() > (tmr+8000)) {
      break;
    }
  }

  // Switch to next game mode
  return GAME_WAITING;
}

// ===================================================================
// == GAME WAITING - PRESS BTNS TO BEGIN =============================
// ===================================================================
int wait_for_players() {
  // show something on the screen to tell players to both hit their buttons
  Serial.println("GAME WAITING...");

  // Params for this message
  scrollAlign = PA_CENTER;
  scrollSpeed = 50;
  scrollPause = 2000; // in ms
  scrollEffectIn = PA_WIPE;
  scrollEffectOut = PA_WIPE;

  // Upper zone; "PRESS"
  // Lower zone: "TO BEGIN",
  P.displayZoneText(ZONE_UPPER, msg[2], scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);
  P.displayZoneText(ZONE_LOWER, msg[3], scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);
  P.displayClear();

  // Display text and loop until both players press buttons
  Serial.println("Waiting for players to press buttons...");
  while(true){
    if (P.displayAnimate()) {
      P.displayReset();
    }
    if (digitalRead(player1pin) == HIGH && digitalRead(player2pin) == HIGH) {
      break;
    }
  }

  Serial.println("Switch to GAME_START...");
  return GAME_START;
}

// ===================================================================
// == START ==========================================================
// ===================================================================
int tell_players_its_on_like_donkey_kong() {
  Serial.println("Game mode message...");

  // Params for text display
  scrollAlign = PA_CENTER;
  scrollSpeed = 50;
  scrollPause = 4000; // in ms
  scrollEffectIn = PA_PRINT;
  scrollEffectOut = PA_WIPE;

  // Upper zone: "FIRST TO"
  // Lower zone: "21 WINS!"
  P.displayZoneText(ZONE_UPPER, msg[4], scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);
  P.displayZoneText(ZONE_LOWER, msg[5], scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);

  // Display text and loop for 4 seconds
  unsigned long tmr = millis();
  while(true){
    if (P.displayAnimate()) {
      P.displayReset();
    }
    if (millis() > (tmr+4000)) {
      break;
    }
  }

  Serial.println("Switch to GAME_ON...");
  P.displayClear();
  return GAME_ON;
}


// ===================================================================
// == CHECK FOR WINNER =]=============================================
// ===================================================================
int there_is_a_winner() {
  int p1 = players[0].score;
  int p2 = players[1].score;

  int score_difference = p1 - p2;
  return ((p1 >= 21 || p2 >= 21) && (score_difference >= 2 || score_difference <= -2));
}


// ===================================================================
// == GAME ON ========================================================
// ===================================================================
int game_on() {

  update_score(players+0);
  update_score(players+1);

  display_score();

  if (there_is_a_winner()) {
    return GAME_OVER;
  } else {
    return GAME_ON;
  }
}


// ===================================================================
// == UPDATE SCORE ===================================================
// ===================================================================
void update_score(Player *player) {

  //  Read the button's state
  int currentButtonState = digitalRead(player->pin);

  if (player->buttonInputState != currentButtonState) {
    player->lastDebounceTime = millis();
  }

  player->buttonInputState = currentButtonState;

  // If past current debounce delay and button pressed,
  // increment score for specified player
  if ((millis() - player->lastDebounceTime) > DEBOUNCE_DELAY) {
    if (player->buttonInputState == HIGH) {
      if (player->buttonState == UNPRESSED) {
        player->buttonState = PRESSED;
        player->score += 1;
      }
    }

    if (player->buttonInputState == LOW) {
      if (player->buttonState == PRESSED) {
        player->buttonState = UNPRESSED;
      }
    }
  } // if millis()
}

// ===================================================================
// == DISPLAY SCORE ==================================================
// ===================================================================
void display_score() {
  if (scoreboard.p1score != players[0].score) {
    scoreboard.p1score = players[0].score;
    redraw_scores();
  }

  if (scoreboard.p2score != players[1].score) {
    scoreboard.p2score = players[1].score;
    redraw_scores();
  }
}

// ===================================================================
// == REDRAW SCORES ==================================================
// ===================================================================
void redraw_scores() {

  int p1 = players[0].score;
  int p2 = players[1].score;
  
  String scores = "";
  scores += p1;
  scores += "-";
  scores += p2;

  // My understanding of C++ is incomplete at best, so i'm likely doing
  // this horribly inefficiently. But "done" is better than "perfect" ;)
  char data[32];
  scores.toCharArray(data, 32);
  strcpy(curMessage, data);
  Serial.println(curMessage);

  // Set font to custom double-high font. Each display shows the top- or
  // bottom-half of the message.
  P.setFont(ZONE_UPPER, fontTallNumTop);
  P.setFont(ZONE_LOWER, fontTallNumBot);

  // Show current score (eg. 17-3) in double-high numerals
  P.displayZoneText(ZONE_UPPER, curMessage, PA_CENTER, 0, 0, PA_PRINT, PA_NO_EFFECT);
  P.displayZoneText(ZONE_LOWER, curMessage, PA_CENTER, 0, 0, PA_PRINT, PA_NO_EFFECT);

  // Delay for short amount of time so text won't strobe
  unsigned long tmr = millis();
  while(true){
    if (P.displayAnimate()) {
      P.displayReset();
    }
    if (millis() > (tmr+100)) {
      break;
    }
  }
}

int game_over() {
  
  // Reset zones to single-high font
  P.setFont(ZONE_UPPER, fontSlim);
  P.setFont(ZONE_LOWER, fontSlim);
  P.setCharSpacing(P.getCharSpacing() * 1); // single height

  // Params for message
  scrollAlign = PA_CENTER;
  scrollSpeed = 50;
  scrollPause = 4000; // in ms
  scrollEffectIn = PA_WIPE;
  scrollEffectOut = PA_WIPE;
  
  // Show winner

  // Again, the C++ is bad, don't judge me (or submit improvement!)
  String bufferr = "";
  if (players[0].score > players[1].score) {
    bufferr = "P1 is the";
  } else {
    bufferr = "P2 is the";
  }

  char data[32];
  bufferr.toCharArray(data, 32);
  strcpy(curMessage, data);
  Serial.println(curMessage);

  // Upper zone: "P1/P2 is the"
  // Lower zone: "Winner!"
  P.displayZoneText(ZONE_UPPER, curMessage, scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);
  P.displayZoneText(ZONE_LOWER, msg[6], scrollAlign, scrollSpeed, scrollPause, scrollEffectIn, scrollEffectOut);

  // Loop for 10 second delay
  unsigned long tmr = millis();
  while(true){
    if (P.displayAnimate()) {
      P.displayReset();
    }
    if (millis() > (tmr+10000)) {
      break;
    }
  }

  // Reset scores, and start over
  players[0].score = 0;
  players[1].score = 0;
   
  return GAME_WAITING;
}

// ===================================================================
// == LOOP ===========================================================
// ===================================================================
void loop() {

  switch(game_state) {
    case SCOREBOARD_INIT :
      game_state = scoreboard_init();
    case GAME_WAITING :
      game_state = wait_for_players();
      break;
    case GAME_START :
      game_state = tell_players_its_on_like_donkey_kong();
      break;
    case GAME_ON :
      game_state = game_on();
      break;
    case GAME_OVER :
      game_state = game_over();
      break;
  }

}
